
package BCE;

import java.util.*;
import java.lang.Math;

public class Graph<K>{
    
    /*
    * Graph represented by adjacency list
    */
    private HashMap<Vertex<K>,Set<Vertex<K>>> adjList;
    
    /*
    * Map of vertex values for convenience
    */
    private HashMap<K,Vertex<K>> vertexValues;
    
    /*
    * Constructor for empty Graph
    */
    public Graph(){
        adjList = new HashMap();
        vertexValues = new HashMap();
    }
    
    
    /*
    * Constructor for Graph without edges
    */
    public Graph(Set<Vertex<K>> V){
        adjList = new HashMap();
        vertexValues = new HashMap();
        for (Vertex<K> v : V){
            adjList.put(v, Collections.emptySet());
            K value = v.getValue();
            vertexValues.put(value,v);
        }
    }

    
    /*
    * Constructor for Graph with given adjacency list
    */
    public Graph(HashMap<Vertex<K>,Set<Vertex<K>>> adj){
        adjList = new HashMap();
        vertexValues = new HashMap();
        adjList = adj;
        for (Vertex<K> v : adj.keySet()){
            vertexValues.put(v.getValue(),v);            
        }
    }
    
     
    /*
    * Input: -
    * Returns: Set of vertices
    */
    public Set<Vertex<K>> getVertices(){
        return adjList.keySet();
    }
    
    /*
    * Input: -
    * Returns: Set of vertices
    */
    public Set<K> getVertexValues(){
        return vertexValues.keySet();
    }
    
    
    /*
    * Input: Value of a vertex
    * Returns: Vertex with the value 
    */
    public Vertex<K> getVertex(K value){
        if(this.getVertexValues().contains(value)){
            return vertexValues.get(value);
        }
        else throw new IllegalArgumentException("Unknown Vertex");
    }
    
    
    /*
    * Input: -
    * Returns: Number of vertices
    */
    public Integer getVertexCount(){
        return adjList.keySet().size();
    }
    
    
    /*
    * Input: -
    * Returns: Set of black vertices
    */   
    public Set<Vertex<K>> getBlackVertices(){
        Set<Vertex<K>> blackSet = new HashSet<>();
        Set<Vertex<K>> vertices = this.getVertices();
        for(Vertex<K> v:vertices){
            if(v.isBlack()){
                blackSet.add(v);
            }
        }
        return blackSet;
    }
    
    /*
    * Input: -
    * Returns: Set of white vertices
    */
    public Set<Vertex<K>> getWhiteVertices(){
        Set<Vertex<K>> whiteSet = new HashSet<>();
        Set<Vertex<K>> vertices = this.getVertices();
        for(Vertex<K> v:vertices){
            if(v.isWhite()){
                whiteSet.add(v);
            }
        }
        return whiteSet;
    }
    
    /*
    * Input: -
    * Returns: Copy of the adjacency list
    */
    public HashMap<Vertex<K>,Set<Vertex<K>>> getAdjListCopy (){
        HashMap<Vertex<K>,Set<Vertex<K>>> copy = new HashMap();
        copy.putAll(this.adjList);   
        return copy;
    }
    
    /*
    * Input: Vertex 
    * Returns: Open neighborhood of input vertex
    */
    public Set<Vertex<K>> getNeighbors(Vertex<K> v){
        if(!adjList.containsKey(v)){
            throw new IllegalArgumentException("Unknown Vertex");
        }
        return adjList.get(v);
    }
    
    /*
    * Input: Vertex 
    * Returns: Open neighborhood of input vertex
    */
    public Set<Vertex<K>> getNeighbors(K x){
        if(!vertexValues.containsKey(x)){
            throw new IllegalArgumentException("Unknown Vertex");
        }
        Vertex<K> v = vertexValues.get(x);
        return adjList.get(v);
    }
    
    
    
    /*
    * Input: Set of vertices
    * Returns: Open neighborhood of input set
    */
    public Set<Vertex<K>> getNeighbors(Set<Vertex<K>> vertices){
        if(!adjList.keySet().containsAll(vertices)){
            throw new IllegalArgumentException("Unknown Vertex");
        }
        Set<Vertex<K>> neighbors = new HashSet<>();
        for(Vertex<K> v : vertices){
            neighbors.addAll(adjList.get(v));
        }
        neighbors.removeAll(vertices);
        return neighbors;
    }
    
    /*
    * Input: -
    * Returns: Set of edges.
    */
    public Set<Edge<K>> getEdges(){
        Set<Vertex<K>> vertices = this.getVertices();
        Set<Vertex<K>> alreadyDone = new HashSet<>();
        Set<Edge<K>> edges = new HashSet<>();
        for(Vertex<K> v : vertices){
            Set<Vertex<K>> neighbors = this.getNeighbors(v);
            alreadyDone.add(v);
            for(Vertex<K> w : neighbors){
                if(!alreadyDone.contains(w)){
                    edges.add(new Edge(v,w));
                }
            }
        }
        return edges;
    }
    
    /*
    * Input: Two sets of vertices
    * Returns: Set of edges between vertices of the two input sets.
    */
    public Set<Edge<K>> getEdges(Set<Vertex<K>> vertexSet1, Set<Vertex<K>> vertexSet2){
        if(!(this.getVertices().containsAll(vertexSet1)&&this.getVertices().containsAll(vertexSet1))){
            throw new IllegalArgumentException("unknown Vertex");
        }
        Set<Edge<K>> edgesBetweenSets = new HashSet<>();
        for(Vertex<K> v : vertexSet1){
            for(Vertex<K> w : vertexSet2){
                if(this.areNeighbors(v, w)){
                    edgesBetweenSets.add(new Edge(v,w));
                }
            }
        }
        return edgesBetweenSets;
    }
      
    
    /*
    * Input: -
    * Returns: Set of missing edges.
    */
    public Set<Edge<K>> getMissingEdges(){
        Set<Vertex<K>> vertices = this.getVertices();
        Set<Vertex<K>> alreadyDone = new HashSet<>();
        Set<Edge<K>> missingEdges = new HashSet<>();
        for(Vertex<K> v : vertices){
            Set<Vertex<K>> neighbors = this.getNeighbors(v);
            alreadyDone.add(v);
            for(Vertex<K> w : vertices){
                if(!alreadyDone.contains(w)){
                    if(!neighbors.contains(w))
                    missingEdges.add(new Edge(v,w));
                }
            }
        }
        return missingEdges;
    }
    
    
    /*
    * Input: -
    * Returns: Set of open P3s
    */
    public Set<Set<Edge<K>>> getP3s(){
        Set<Vertex<K>> vertices = this.getVertices();
        Set<Set<Edge<K>>> p3s = new HashSet<>();
        for(Vertex<K> v : vertices){
            Set<Vertex<K>> alreadyDone = new HashSet<>();
            Set<Vertex<K>> neighbors = this.getNeighbors(v);
            for(Vertex<K> w : neighbors){
                alreadyDone.add(w);
                for(Vertex<K> u : neighbors){
                    if(!alreadyDone.contains(u)){
                        if(!(this.areNeighbors(w,u))){
                            
                            Set<Edge<K>> temp = new HashSet<>();
                            temp.add(new Edge(u,v));
                            temp.add(new Edge(v,w));
                            p3s.add(temp);
                        }
                    }
                }
            }
        }
        return p3s;
    }
    
    /*
    * Input: -
    * Returns: Set of open P3s as set of vertices
    */
    public Set<Set<Vertex<K>>> getP3Vertices(){
        Set<Vertex<K>> vertices = this.getVertices();
        Set<Set<Vertex<K>>> p3s = new HashSet<>();
        for(Vertex<K> v : vertices){
            Set<Vertex<K>> alreadyDone = new HashSet<>();
            Set<Vertex<K>> neighbors = this.getNeighbors(v);
            for(Vertex<K> w : neighbors){
                alreadyDone.add(w);
                for(Vertex<K> u : neighbors){
                    if(!alreadyDone.contains(u)){
                        if(!(this.areNeighbors(w,u))){
                            Set<Vertex<K>> temp = new HashSet<>();
                            temp.add(u);
                            temp.add(v);
                            temp.add(w);
                            p3s.add(temp);
                        }
                    }
                }
            }
        }
        return p3s;
    }
    
    
    /*
    * Input: -
    * Returns: Set of triangles in the graph
    */
    public Set<Set<Edge<K>>> getTriangles(){
        Set<Vertex<K>> vertices = this.getVertices();
        Set<Vertex<K>> finished = new HashSet<>();
        Set<Set<Edge<K>>> triangles = new HashSet<>();
        for(Vertex<K> v : vertices){
            finished.add(v);
            Set<Vertex<K>> alreadyDone = new HashSet<>();
            Set<Vertex<K>> neighbors = this.getNeighbors(v);
            for(Vertex<K> w : neighbors){
                if(!finished.contains(w)){
                    alreadyDone.add(w);
                    for(Vertex<K> u : neighbors){
                        if(!finished.contains(u)){
                            if(!alreadyDone.contains(u)){
                                if(this.areNeighbors(w,u)){
                                    Edge<K> edge1 = new Edge(u,v);
                                    Edge<K> edge2 = new Edge(v,w);
                                    Edge<K> edge3 = new Edge(u,w);
                                    Set<Edge<K>> temp = new HashSet<>();
                                    temp.add(edge1);
                                    temp.add(edge2);
                                    temp.add(edge3);
                                    triangles.add(temp);
                                }
                            }
                        }
                    }                   
                }
            }
        }
        return triangles;
    }
        
    /*
    * Input: Vertex to be added
    * Returns: True if vertex was added (if vertex didn't exist already), false otherwise
    */
    public boolean addVertex(Vertex<K> v){
        if(adjList.containsKey(v)){
            return false;
        } else {
            adjList.put(v, Collections.emptySet());
            vertexValues.put(v.getValue(), v);
            return true;
        }
    }
    
    /*
    * Input: Set of vertices to be added
    * Returns: True if vertices were added (if every vertex didn't exist already), false otherwise
    */
    public boolean addVertices(Set<Vertex<K>> V){
        for(Vertex<K> v:V){
            if(adjList.containsKey(v)){
                return false;
            }
        }
        for(Vertex<K> v:V){
                adjList.put(v, Collections.emptySet());
                vertexValues.put(v.getValue(), v);
        }
        return true;
    }    
    
    
    /*
    * Input: Value of vertex to be added
    * Returns: True if vertex was added (if vertex didn't exist already), false otherwise
    */
    public boolean addVertex(K x, String color){
        Vertex<K> v = new Vertex(x,color);
        if(adjList.containsKey(v)){
            return false;
        } else {
            adjList.put(v, Collections.emptySet());
            vertexValues.put(v.getValue(), v);
            return true;
        }
    }
    
    /*
    * Input: An edge
    * Returns: True if edge was added, false if not.
    */
    public boolean addEdge(Edge<K> e){
        Vertex<K>[] vertices = e.getNodeArray();
        Vertex<K> v1 = vertices[0];
        Vertex<K> v2 = vertices[1];
        if(v1 == v2){
            throw new IllegalArgumentException("Loops are not allowed.");
        }
        if(!adjList.containsKey(v1) || !adjList.containsKey(v2)){
            throw new IllegalArgumentException("unknown Vertex");
        }
        Set<Vertex<K>> neighbors_v1 = new HashSet<>();
        neighbors_v1.addAll(this.getNeighbors(v1));
        Set<Vertex<K>> neighbors_v2 = new HashSet<>();
        neighbors_v2.addAll(this.getNeighbors(v2));
        boolean newEdge = neighbors_v1.add(v2) && neighbors_v2.add(v1);
        adjList.put(v1,neighbors_v1);
        adjList.put(v2,neighbors_v2);
        return newEdge;
    }   
    
    
    /*
    * Input: Two vertices
    * Returns: True if edge was added, false if not.
    */
    public boolean addEdge(Vertex<K> v1, Vertex<K> v2){
        if(v1 == v2){
            throw new IllegalArgumentException("Loops are not allowed.");
        }
        if(!adjList.containsKey(v1) || !adjList.containsKey(v2)){
            throw new IllegalArgumentException("unknown Vertex");
        }
        Set<Vertex<K>> neighbors_v1 = new HashSet<>();
        neighbors_v1.addAll(this.getNeighbors(v1));
        Set<Vertex<K>> neighbors_v2 = new HashSet<>();
        neighbors_v2.addAll(this.getNeighbors(v2));
        boolean newEdge = neighbors_v1.add(v2) && neighbors_v2.add(v1);
        adjList.put(v1,neighbors_v1);
        adjList.put(v2,neighbors_v2);
        return newEdge;
    }
    
    /*
    * Input: Two values of vertices
    * Returns: True if edge was added, false if not.
    */
    public boolean addEdge(K x, K y){
        if(x == y){
            throw new IllegalArgumentException("Loops are not allowed.");
        }
        if(!vertexValues.containsKey(x) || !vertexValues.containsKey(y)){
            throw new IllegalArgumentException("unknown Vertex");
        }
        Vertex<K> v1 = vertexValues.get(x);
        Vertex<K> v2 = vertexValues.get(y);
        Set<Vertex<K>> neighbors_v1 = new HashSet<>();
        neighbors_v1.addAll(this.getNeighbors(v1));
        Set<Vertex<K>> neighbors_v2 = new HashSet<>();
        neighbors_v2.addAll(this.getNeighbors(v2));
        boolean newEdge = neighbors_v1.add(v2) && neighbors_v2.add(v1);
        adjList.put(v1,neighbors_v1);
        adjList.put(v2,neighbors_v2);
        return newEdge;
    }
    
    
    /*
    * Input: Adds all edges between two sets of vertices
    * Returns: True if all edges were succesfully added, false if not.
    */
    public boolean addAllEdges(Set<Vertex<K>> V1, Set<Vertex<K>> V2){
        Boolean allWell = true;
        for(Vertex<K> v1 : V1){
            for(Vertex<K> v2 : V2){
                if(v1 != v2){
                    if(!adjList.containsKey(v1) || !adjList.containsKey(v2)){
                        allWell = false;
                    } else {
                        this.addEdge(v1, v2);
                    }                    
                }  else {
                    allWell = false;
                } 
            }
        }
        return allWell;
    }
    
    
    /*
    * Input: Vertex to be deleted from the graph
    * Returns: True if vertex was successfully deleted
    */
    public boolean deleteVertex(Vertex<K> v){
        if(!adjList.containsKey(v)){
            return false;
        } else {
            Set<Vertex<K>> neighbors = this.getNeighbors(v);
            for(Vertex<K> u : neighbors){
                Set<Vertex<K>> NeighborsOfNeighbor = this.getNeighbors(u);
                NeighborsOfNeighbor.remove(v);
                adjList.put(u,NeighborsOfNeighbor);
            }
            K value = v.getValue();
            adjList.remove(v);
            vertexValues.remove(value);
            return true;
        }
    }
    
    /*
    * Input: Vertex to be deleted from the graph
    * Returns: True if vertex was successfully deleted
    */
    public boolean deleteVertexByValue(K value){
        if(!vertexValues.containsKey(value)){
            return false;
        } else {
            Vertex<K> v = this.vertexValues.get(value);
            Set<Vertex<K>> neighbors = this.getNeighbors(v);
            for(Vertex<K> u : neighbors){
                Set<Vertex<K>> NeighborsOfNeighbor = this.getNeighbors(u);
                NeighborsOfNeighbor.remove(v);
                adjList.put(u,NeighborsOfNeighbor);
            }
            adjList.remove(v);
            vertexValues.remove(value);
            return true;
        }
    }
    
    /*
    * Input: Edge to be deleted from the graph
    * Returns: True if edge was successfully deleted
    */
    public boolean deleteEdge(Edge<K> e){
        Vertex<K>[] vertices = e.getNodeArray();
        Vertex<K> v = vertices[0];
        Vertex<K> u = vertices[1];
        
        if(!adjList.containsKey(v) || !adjList.containsKey(u) ){
            return false;
        } else {
            Set<Vertex<K>> neighborsU = this.getNeighbors(u);
            neighborsU.remove(v);
            adjList.put(u,neighborsU);
            Set<Vertex<K>> neighborsV = this.getNeighbors(v);
            neighborsV.remove(u);
            adjList.put(v,neighborsV);
            return true;
        }
    }
    
    /*
    * Input: Edge to be deleted from the graph
    * Returns: True if edge was successfully deleted
    */
    public boolean deleteEdgeByValue(K value1, K value2){
        if(!vertexValues.containsKey(value1) || !vertexValues.containsKey(value2) ){
            return false;
        }
        Vertex<K> v = vertexValues.get(value1);
        Vertex<K> u = vertexValues.get(value2);       
        if(!adjList.get(v).contains(u) || !adjList.get(u).contains(v) ){
            return false;
        } else {
            Set<Vertex<K>> neighborsU = this.getNeighbors(u);
            neighborsU.remove(v);
            adjList.put(u,neighborsU);
            Set<Vertex<K>> neighborsV = this.getNeighbors(v);
            neighborsV.remove(u);
            adjList.put(v,neighborsV);
            return true;
        }
    }
    
    /*
    * Input: Vertex to get degree of
    * Returns: Degree of vertex
    */
    public int getVertexDegree(Vertex<K> v){
        Set<Vertex<K>> neighbours = this.getNeighbors(v);
        return neighbours.size();
    }
    
    /*
    * Input: -
    * Returns: List of degrees of all vertices
    */
    public List<Integer> getVertexDegrees(){
        List<Integer> degrees = new ArrayList<>();
        Set<Vertex<K>> vertices = this.getVertices();
        for(Vertex<K> v:vertices){
            int deg = this.getVertexDegree(v);
            degrees.add(deg);
        }
        return degrees;
    }
    
    /*
    * Inpute: -
    * Returns: HashMap of all vertices mapped to their degrees
    */
    public HashMap<Vertex<K>,Integer> getVertexDegreesMap(){
        HashMap degrees = new HashMap<>();
        Set<Vertex<K>> vertices = this.getVertices();
        for(Vertex<K> v:vertices){
            int deg = this.getVertexDegree(v);
            degrees.put(v,deg);
        }
        return degrees;    
    }
    
    
   
    
    /*
    * Input: Set of Vertices
    * Returns: Graph induced by input set of vertices
    */
    public Graph getInducedSubgraph(Set<Vertex<K>> vertices){
        HashMap<Vertex<K>,Set<Vertex<K>>> inducedAdjList = new HashMap();
        for(Vertex<K> v : vertices){
            Set<Vertex<K>> neighbors = new HashSet<>();
            neighbors.addAll(this.getNeighbors(v));
            neighbors.retainAll(vertices);
            inducedAdjList.put(v,neighbors);
        }
        return new Graph(inducedAdjList);
    }
    
    /*
    * Implements a breadth first search on the graph
    * Input: Start vertex
    * Returns: Set of vertices found by the BFS
    */
    public Set<Vertex<K>> BFS(Vertex<K> start){
        Set<Vertex<K>> discovered = new HashSet<>();
        Queue<Vertex<K>> q = new LinkedList<>();
        q.offer(start);
        discovered.add(start);
        while (!q.isEmpty())
        {
            Vertex <K> v = q.remove();               
            for (Vertex<K> i : getNeighbors(v))
            {
                if (!discovered.contains(i))
                {
                    q.offer(i);
                    discovered.add(i);
                }
            }
        }
        return discovered;
    }
    
    /*
    * Input: Start vertex
    * Returns: Set of vertices of the connected component containing start vertex
    */
    public Set<Vertex<K>> getConnectedComponentVertices(Vertex<K> start){
        Set<Vertex<K>> vertexSet = this.BFS(start);
        return vertexSet;
    }    
    
    /*
    * Input: Start vertex
    * Returns: Graph of the connected component containing start vertex
    */
    public Graph getConnectedComponent(Vertex<K> start){
        Set<Vertex<K>> vertexSet = this.BFS(start);
        Graph connectedComponent = this.getInducedSubgraph(vertexSet);
        return connectedComponent;
    }
    
    /*
    * Input: -
    * Returns: Set of connected components
    */
    public Set<Graph> getConnectedComponents(){
        Set<Graph> components = new HashSet();
        Set<Vertex<K>> vertices = this.getVertices();
        Set<Vertex<K>> vertexSet = new HashSet();
        vertexSet.addAll(vertices);     
        Set<Vertex<K>> componentVertices;
        while(!vertexSet.isEmpty()){
            Vertex<K> v = vertexSet.iterator().next();
            componentVertices = this.getConnectedComponentVertices(v);
            Graph connectedComponent = getInducedSubgraph(componentVertices);
            components.add(connectedComponent);
            vertexSet.removeAll(componentVertices);
        }        
        return components;
    }
    
    
    /*
    * Input: Set of Vertices
    * Returns: True if input set is a clique, false otherwise
    */
    public boolean isClique(Set<Vertex<K>> vertices){
        if(!this.getVertices().containsAll(vertices)){
            throw new IllegalArgumentException("unknown Vertices");
        }
        boolean isClique = true;
        Set<Vertex<K>> alreadyDone = new HashSet<>();
        for(Vertex<K> v : vertices){
            alreadyDone.add(v);
            for(Vertex<K> w : vertices){
                if(!alreadyDone.contains(w)){
                    isClique = isClique && this.areNeighbors(v, w);
                }
            }
        }
        return isClique;
    }
    
    
    /*
    * Input: -
    * Returns: True if graph is a clique, false otherwise
    */
    public boolean isCluster(){
        Set<Vertex<K>> vertices = this.getVertices();
        boolean isCluster = true;
        Set<Vertex<K>> alreadyDone = new HashSet<>();
        for(Vertex<K> v : vertices){
            alreadyDone.add(v);
            for(Vertex<K> w : vertices){
                if(!alreadyDone.contains(w)){
                    if(!this.areNeighbors(v, w)){
                        return false;
                    }
                }
            }
        }
        return isCluster;
    }
    
    /*
    * Input: -
    * Returns: True if graph is a singleton, false otherwise
    */
    public boolean isSingleton(){
        if (this.getVertices().size()==1){
            return true;
        }
        return false;
    }
    
    /*
    * Input: -
    * Returns: True if graph is a single edge, false otherwise
    */
    public boolean isIsolatedEdge(){
        if (this.getVertices().size() == 2){
            if (this.getEdges().size()== 1){
                return true;
            }
        }
        return false;   
    }
    
    
    // Assumption: Graph is a cluster
    /*
    * Input: -
    * Returns: True if graph is a monochromatic black cluster, false otherwise
    */
    public boolean isMonochromaticBlackCluster(){
        Set<Vertex<K>> vertices = this.getVertices();
        Set<Vertex<K>> blackVertices = this.getBlackVertices();
        if(vertices.equals(blackVertices)){
            return true;
        }
        return false;
    }
    
    // Assumption: Graph is a cluster
    /*
    * Input: -
    * Returns: True if graph is a monochromatic white cluster, false otherwise
    */
    public boolean isMonochromaticWhiteCluster(){
        Set<Vertex<K>> vertices = this.getVertices();
        Set<Vertex<K>> whiteVertices = this.getWhiteVertices();
        if(vertices.equals(whiteVertices)){
            return true;
        }
        return false;
    }
    
    // Assumption: Graph is a cluster
    /*
    * Input: -
    * Returns: True if graph is a valid black-dominated cluster, false otherwise
    */
    public boolean isValidBlackDominatedCluster(){
        Set<Vertex<K>> blackVertices = this.getBlackVertices();
        Set<Vertex<K>> whiteVertices = this.getWhiteVertices();
        if(whiteVertices.size() == 1 && blackVertices.size() >= 1){
            return true;
        }
        return false;
    }
    
    // Assumption: Graph is a cluster#
    /*
    * Input: -
    * Returns: True if graph is a valid white-dominated cluster, false otherwise
    */
    public boolean isValidWhiteDominatedCluster(){
        Set<Vertex<K>> blackVertices = this.getBlackVertices();
        Set<Vertex<K>> whiteVertices = this.getWhiteVertices();
        if(blackVertices.size() == 1 && whiteVertices.size() >= 1){
            return true;
        }
        return false;
    }
    
    
    /*
    * Input: Two vertices to be checked whether they are adjacent
    * Returns: True if vertices are adjacent, false otherwise
    */
    public boolean areNeighbors(Vertex<K> v1, Vertex<K> v2){
       if(!this.getVertices().contains(v1) || !this.getVertices().contains(v2) ){
            throw new IllegalArgumentException("unknown Vertices");
        } 
       boolean areNeighbors = false;
       if(this.getNeighbors(v1).size() > this.getNeighbors(v2).size()){
           areNeighbors = this.getNeighbors(v2).contains(v1);
       } else {
           areNeighbors = this.getNeighbors(v1).contains(v2);
       }
       return areNeighbors;
    }
    
    
    
    /*
    * Input: -
    * Returns: A P3 if one exists, an empty set otherwise
    */
    public Set<Edge<K>> isP3Free(){
        Set<Vertex<K>> vertices = this.getVertices();
        Set<Edge<K>> noP3 = new HashSet<>();
        for(Vertex<K> v : vertices){
            Set<Vertex<K>> alreadyDone = new HashSet<>();
            Set<Vertex<K>> neighbors = this.getNeighbors(v);
            for(Vertex<K> w : neighbors){
                alreadyDone.add(w);
                for(Vertex<K> u : neighbors){
                    if(!alreadyDone.contains(u)){
                        if(!(this.areNeighbors(w,u))){
                            Edge<K> edge1 = new Edge(u,v);
                            Edge<K> edge2 = new Edge(v,w);
                            Set<Edge<K>> P3 = new HashSet<>();
                            P3.add(edge1);
                            P3.add(edge2);
                            return P3;
                        }
                    }
                }
            }
        }
        return noP3;
    }  
    
    /*
    * Input: -
    * Returns: A set of connected vertices with conflicting colors if one exists, an empty set otherwise
    */
    public Set<Vertex<K>> noColorConflicts(){
        Set<Vertex<K>> vertices = this.getVertices();
        Set<Vertex<K>> noConflict = new HashSet<>();
        for(Vertex<K> v : vertices){
            String mainColor = v.getColor();
            int mainColorCount = 1;
            int secondColorCount = 0;
            Set<Vertex<K>> conflicts = new HashSet<>();
            Set<Vertex<K>> neighbors = this.getNeighbors(v);
            for(Vertex<K> w : neighbors){
                if(mainColorCount<2 || secondColorCount<2){
                    if(w.getColor().equals(mainColor)){
                        if(mainColorCount<2){
                            conflicts.add(w);
                            mainColorCount++;
                        }
                    } else {
                        if(secondColorCount<2){
                            conflicts.add(w);
                            secondColorCount++;
                        }
                    }
                } else {
                    conflicts.add(v);
                    return conflicts;
                }
            }
            if(!(mainColorCount<2 || secondColorCount<2)){
                conflicts.add(v);
                return conflicts;
            }
        }
        return noConflict;
    }
    
    
    /*
    * Input: -
    * Returns: All sets of connected vertices with conflicting colors if one exists, an empty set otherwise
    */
    public Set<Set<Vertex<K>>> allColorConflicts(){
        Set<Vertex<K>> vertices = this.getVertices();
        Set<Set<Vertex<K>>> conflictSets = new HashSet<>();
        for(Vertex<K> v : vertices){
            Set<Vertex<K>> neighbors = this.getNeighbors(v);
            for(Vertex<K> w : neighbors){
                for(Vertex<K> x : neighbors){
                    if(x==w){
                        continue;
                    }
                    for(Vertex<K> y : neighbors){
                        if(y==x || y==w){
                            continue;
                        }
                        int blackColorCount = 0;
                        int whiteColorCount = 0;
                        Set<Vertex<K>> candidates = new HashSet<>();
                        candidates.add(v);                              
                        candidates.add(w);
                        candidates.add(x);
                        candidates.add(y);
                        if(this.isClique(candidates)){
                            for(Vertex<K> u : candidates){
                                if(u.isBlack()){
                                    blackColorCount++;
                                }
                                else whiteColorCount++;
                            }
                            if(blackColorCount==2 && whiteColorCount==2){
                                if(!conflictSets.contains(candidates)){
                                    conflictSets.add(candidates);
                                } 
                            } 
                        }  
                    }
                }
            }      
        }
      //  System.out.println(conflictSets);
        return conflictSets;
    }    
    
        
    
    /*
    * Input: -
    * Returns: True, if a Vertex with degree 0 exists, false otherwise
    */
    public Boolean hasSingleton(){
        Set<Vertex<K>> vertices = this.getVertices();
        for(Vertex<K> v : vertices){
            if(this.getVertexDegree(v)==0){
                return true;
            }
        }
        return false;
    }   
    
    /*
    * Input: -
    * Returns: A Vertex with degree 0, if one exists
    */
    public Vertex<K> returnSingleton(){
        Set<Vertex<K>> vertices = this.getVertices();
        for(Vertex<K> v : vertices){
            if(this.getVertexDegree(v)==0){
                return v;
            }
        }
        return null;
    }
    
    
    /*
    * Input: Number of connectedcomponents, maximum number of vertices per component, seed for random generator
    * Returns: Test graph
    */
    public Graph<Integer> createTestGraph(int connectedComponents, int vertexCount, long seed){
        int i = vertexCount;
        int j = connectedComponents;
        int count = 1;
        Random generator = new Random(seed);
        Graph<Integer> resultGraph = new Graph<>();
        while(j>0){
            int numberOfVertices = generator.nextInt(i)+1;            
            int typeOfCluster = generator.nextInt(4)+1;
            if(numberOfVertices == 1){
                typeOfCluster = generator.nextInt(2)+1;
            }                     
            Set<Vertex<Integer>> vertexSet = new HashSet<>();
            switch(typeOfCluster){
                case 1:
                    System.out.println("create monochromatic black cluster of size "+numberOfVertices);
                    for(int x = count; x < count+numberOfVertices; x++){
                        Vertex<Integer> newVertex = new Vertex<>(x,"b");
                        vertexSet.add(newVertex);
                    }
                    resultGraph.addVertices(vertexSet);
                    resultGraph.addAllEdges(vertexSet, vertexSet);
                    count += numberOfVertices;
                    break;
                case 2:
                    System.out.println("create monochromatic white cluster of size "+numberOfVertices);
                    for(int x = count; x < count+numberOfVertices; x++){
                        Vertex<Integer> newVertex = new Vertex<>(x,"w");
                        vertexSet.add(newVertex);
                    }
                    resultGraph.addVertices(vertexSet);
                    resultGraph.addAllEdges(vertexSet, vertexSet);
                    count += numberOfVertices;
                    break;
                case 3:
                    System.out.println("create black-dominated cluster of size "+numberOfVertices);
                    for(int x = count; x < count+numberOfVertices-1; x++){
                        Vertex<Integer> newVertex = new Vertex<>(x,"b");
                        vertexSet.add(newVertex);
                    }
                    vertexSet.add(new Vertex<>(count+numberOfVertices-1,"w"));
                    resultGraph.addVertices(vertexSet);
                    resultGraph.addAllEdges(vertexSet, vertexSet);
                    count += numberOfVertices;
                    break;
                case 4:
                    System.out.println("create white-dominated cluster of size "+numberOfVertices);
                    for(int x = count; x < count+numberOfVertices-1; x++){
                        Vertex<Integer> newVertex = new Vertex<>(x,"w");
                        vertexSet.add(newVertex);
                    }
                    vertexSet.add(new Vertex<>(count+numberOfVertices-1,"b"));
                    resultGraph.addVertices(vertexSet);
                    resultGraph.addAllEdges(vertexSet, vertexSet);
                    count += numberOfVertices;
                    break;
                default:
                    System.out.println("wrong type of cluster");
                    break;
            }
            j -= 1;
            //System.out.println(j);
        } 
        return resultGraph;
    }
    
    
    
    /*
    * Input: Number of connectedcomponents, maximum number of vertices per component
    * Returns: Test graph
    */
    public Graph<Integer> createTestGraph(int connectedComponents, int vertexCount){
        int i = vertexCount;
        int j = connectedComponents;
        int count = 1;
        Graph<Integer> resultGraph = new Graph<>();
        while(j>0){
            int numberOfVertices = (int)(Math.random()*i)+1;            
            int typeOfCluster = (int)(Math.random()*4)+1;
            if(numberOfVertices == 1){
                typeOfCluster = (int)(Math.random()*2)+1;
            }                     
            Set<Vertex<Integer>> vertexSet = new HashSet<>();
            switch(typeOfCluster){
                case 1:
                    System.out.println("create monochromatic black cluster of size "+numberOfVertices);
                    for(int x = count; x < count+numberOfVertices; x++){
                        Vertex<Integer> newVertex = new Vertex<>(x,"b");
                        vertexSet.add(newVertex);
                    }
                    resultGraph.addVertices(vertexSet);
                    resultGraph.addAllEdges(vertexSet, vertexSet);
                    count += numberOfVertices;
                    break;
                case 2:
                    System.out.println("create monochromatic white cluster of size "+numberOfVertices);
                    for(int x = count; x < count+numberOfVertices; x++){
                        Vertex<Integer> newVertex = new Vertex<>(x,"w");
                        vertexSet.add(newVertex);
                    }
                    resultGraph.addVertices(vertexSet);
                    resultGraph.addAllEdges(vertexSet, vertexSet);
                    count += numberOfVertices;
                    break;
                case 3:
                    System.out.println("create black-dominated cluster of size "+numberOfVertices);
                    for(int x = count; x < count+numberOfVertices-1; x++){
                        Vertex<Integer> newVertex = new Vertex<>(x,"b");
                        vertexSet.add(newVertex);
                    }
                    vertexSet.add(new Vertex<>(count+numberOfVertices-1,"w"));
                    resultGraph.addVertices(vertexSet);
                    resultGraph.addAllEdges(vertexSet, vertexSet);
                    count += numberOfVertices;
                    break;
                case 4:
                    System.out.println("create white-dominated cluster of size "+numberOfVertices);
                    for(int x = count; x < count+numberOfVertices-1; x++){
                        Vertex<Integer> newVertex = new Vertex<>(x,"w");
                        vertexSet.add(newVertex);
                    }
                    vertexSet.add(new Vertex<>(count+numberOfVertices-1,"b"));
                    resultGraph.addVertices(vertexSet);
                    resultGraph.addAllEdges(vertexSet, vertexSet);
                    count += numberOfVertices;
                    break;
                default:
                    System.out.println("wrong type of cluster");
                    break;
            }
            j -= 1;
            //System.out.println(j);
        } 
        return resultGraph;
    }
    
    /*
    * Input: -
    * Returns: Copy of this graph
    */
    public Graph<K> copyGraph(){
        HashMap<Vertex<K>,Set<Vertex<K>>> newAdjList = new HashMap<>();
        for(Vertex<K> v : this.adjList.keySet()){
            Set<Vertex<K>> temp = new HashSet<>();
            for(Vertex<K> w : this.adjList.get(v)){
                temp.add(w);
            }                
            newAdjList.put(v, temp);
        }
        Graph<K> resultGraph = new Graph(newAdjList);
        return resultGraph;
    }
    
    
    /*
    * Input: Number of distortions, seed for random generator
    * Returns: Copy of this graph with some edges inserted or deleted
    */    
    public Graph<K> distortGraph(int distortions, long seed){
        Graph<K> newGraph = this.copyGraph();
        Random generator = new Random(seed);
        ArrayList<Vertex<K>> vertices = new ArrayList<>(newGraph.adjList.keySet());
        int numberOfVertices = vertices.size();
        Set<Edge<K>> alreadyModified = new HashSet<>();
        while(distortions > 0){
            int numberOfFirstVertex = (int)(generator.nextInt(numberOfVertices));
            int numberOfSecondVertex = (int)(generator.nextInt(numberOfVertices));
            if(numberOfFirstVertex != numberOfSecondVertex){
                Vertex<K> v = vertices.get(numberOfFirstVertex);
                Vertex<K> w = vertices.get(numberOfSecondVertex);
                Edge<K> e = new Edge(v,w);
                if(!alreadyModified.contains(e)){
                    if(newGraph.adjList.get(v).contains(w)){
                        newGraph.deleteEdge(e);
                    } else {
                        newGraph.addEdge(e);
                    }
                    alreadyModified.add(e);
                } 
                distortions -= 1;
            } 
        }
        return newGraph;
    }
    
    
    /*
    * Input: Number of distortions
    * Returns: Copy of this graph with some edges inserted or deleted
    */    
    public Graph<K> distortGraph(int distortions){
        Graph<K> newGraph = this.copyGraph();
        ArrayList<Vertex<K>> vertices = new ArrayList<>(newGraph.adjList.keySet());
        int numberOfVertices = vertices.size();
        Set<Edge<K>> alreadyModified = new HashSet<>();
        while(distortions > 0){
            int numberOfFirstVertex = (int)(Math.random()*numberOfVertices);
            int numberOfSecondVertex = (int)(Math.random()*numberOfVertices);
            if(numberOfFirstVertex != numberOfSecondVertex){
                Vertex<K> v = vertices.get(numberOfFirstVertex);
                Vertex<K> w = vertices.get(numberOfSecondVertex);
                Edge<K> e = new Edge(v,w);
                if(!alreadyModified.contains(e)){
                    if(newGraph.adjList.get(v).contains(w)){
                        newGraph.deleteEdge(e);
                    } else {
                        newGraph.addEdge(e);
                    }
                    alreadyModified.add(e);
                } 
                distortions -= 1;
            } 
        }
        return newGraph;
    }
    
    
    /*
    * Input: Graph G
    * Returns: New Integer graph isomorphic to G
    */    
    public Graph<Integer> transformToIntegerGraph(){
        Set<Vertex<K>> vertices = this.getVertices();
        Set<Edge<K>> edges = this.getEdges();
        Set<Vertex<Integer>> newVertices = new HashSet();
        HashMap<Vertex<K>,Vertex<Integer>> vertexMapping = new HashMap();
        HashMap<Vertex<Integer>,Set<Vertex<Integer>>> newAdjList;
        int vertexCount = 1;
        for(Vertex<K> vertex : vertices){
            Vertex<Integer> newVertex = new Vertex(vertexCount,vertex.getColor());
            vertexMapping.put(vertex, newVertex);
            newVertices.add(newVertex);
            vertexCount++;
        }
        Graph<Integer> newGraph = new Graph(newVertices);
        for(Edge<K> e : edges){
            Vertex<K>[] nodes = e.getNodeArray();
            Vertex<Integer> v1 = vertexMapping.get(nodes[0]);
            Vertex<Integer> v2 = vertexMapping.get(nodes[1]);
            Edge<Integer> newEdge = new Edge(v1,v2);
            newGraph.addEdge(newEdge);
        }
        return newGraph;        
    }
    
    
  
    
    @Override
    public String toString(){
        String output = "Graph:";
        Set<Vertex<K>> vertices = this.getVertices();
        output += vertices.toString() + System.lineSeparator();
        Set<Edge<K>> edges = this.getEdges();
        for(Edge<K> e : edges){
            output += e.toString() + System.lineSeparator();
        }
        return output;
    }
    
   
}

