

package BCE;

import java.io.FileWriter;
import java.io.IOException;
import java.util.*;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.Files;
import java.nio.charset.Charset;
import gurobi.*;
import java.util.concurrent.TimeUnit;

public class ILPs<K> {
    
   
     /*
    * Solves BCD for given graph G and returns a tuple of the solution set and statistics string
    */
    public Tuple<Set<Edge<K>>,String> solveBCD (Graph G) {
       
        try {
            String path =  System.getProperty("java.library.path");
            System.out.println("Java path: " + path);
            
            long StartTime = System.nanoTime();
            
            Set<Edge<K>> edgesInG = G.getEdges(); 
            Set<Vertex<K>> verticesInG = G.getVertices();
            Set<Vertex<K>> blackVerticesInG = G.getBlackVertices();
            Set<Vertex<K>> whiteVerticesInG = G.getWhiteVertices();
            
            //setup
            GRBEnv    env   = new GRBEnv();
            FileWriter logfile = new FileWriter("solveBCD.log");
            GRBModel  model = new GRBModel(env);
            
            model.set(GRB.DoubleParam.TimeLimit, 600);
            model.set(GRB.IntParam.LazyConstraints, 1);
                                    

            
            // Create edge variables
            HashMap<Edge<K>,GRBVar> edgeVariables = new HashMap<>();
            for(Edge<K> e : edgesInG){
               edgeVariables.put(e,model.addVar(0.0, 1.0, 0.0, GRB.BINARY, e.toString())); 
            }
            
            // Set objective
            GRBLinExpr objective = new GRBLinExpr();
            for(GRBVar var : edgeVariables.values()){
                objective.addTerm(-1.0, var);
            }
            objective.addConstant(edgesInG.size());
            model.setObjective(objective, GRB.MINIMIZE);

            
            model.update();
            
            // Add constraints for each P3
            int P3ConstraintCount = 0;
            
            Set<Set<Edge<K>>> p3s = G.getP3s();
            for(Set<Edge<K>> p3 : p3s){
                Edge<K>[] edges = new Edge[2];
                p3.toArray(edges);
                Edge<K> e1 = edges[0];
                Edge<K> e2 = edges[1];
                GRBLinExpr term = new GRBLinExpr();
                term.addTerm(1.0,edgeVariables.get(e1));
                term.addTerm(1.0,edgeVariables.get(e2));
                
//                System.out.println("P3 constraint added: "+ 
//                        term.getCoeff(0) +term.getVar(0).get(GRB.StringAttr.VarName) +" "+
//                        term.getCoeff(1)+term.getVar(1).get(GRB.StringAttr.VarName));
                model.addConstr(term, GRB.LESS_EQUAL, 1.0, p3.toString());  
                P3ConstraintCount ++;
            }
            System.out.println("P3 Constraints added: "+P3ConstraintCount);
            
            
            // Add constraints for each triangle (potential P3)
            int TriangleConstraintCount = 0;
            Set<Set<Edge<K>>> triangles = G.getTriangles();
            for(Set<Edge<K>> tri : triangles){
                for(Edge<K> e : tri){
                    GRBLinExpr term = new GRBLinExpr();
                    for(Edge<K> e2 : tri){
                        if(e2.equals(e)){
                            term.addTerm(-1.0,edgeVariables.get(e2));
                        }
                        else term.addTerm(1.0,edgeVariables.get(e2));                      
                    } 
                    model.addConstr(term, GRB.LESS_EQUAL, 1.0, tri.toString());
                    TriangleConstraintCount ++;
//                    System.out.println("Triangle constraint added: "+ 
//                        term.getCoeff(0)+term.getVar(0).get(GRB.StringAttr.VarName)+" "+
//                        term.getCoeff(1)+term.getVar(1).get(GRB.StringAttr.VarName)+" "+
//                        term.getCoeff(2)+term.getVar(2).get(GRB.StringAttr.VarName));
                }                                 
            }
            System.out.println("Triangle Constraints added: "+TriangleConstraintCount);
            
            
            
            
            //Add lazy constraints via callback
            CallbackBCD cb = new CallbackBCD(model.getVars(), logfile, G, edgeVariables);
            model.setCallback(cb);
            
            
            
            // Optimize model
            model.optimize();
            
            
            
            long FinishTime = System.nanoTime();
            long ILPTime = FinishTime - StartTime;
            long convertedILPtime = TimeUnit.MILLISECONDS.convert(ILPTime, TimeUnit.NANOSECONDS);
            
            String finishedInTime;
            if(model.get(GRB.IntAttr.Status)==(GRB.Status.TIME_LIMIT)){
                finishedInTime = "no";
            }
            else finishedInTime = "yes";
            
            Graph resultGraph = new Graph(verticesInG);
            for(Edge e : edgeVariables.keySet()){
                if((edgeVariables.get(e).get(GRB.DoubleAttr.X))>=0.9999){
                    Vertex<K>[] nodeArray = e.getNodeArray();
                    resultGraph.addEdge(nodeArray[0],nodeArray[1]);
                }
            }
            
            Set<Graph> resultComponents = resultGraph.getConnectedComponents();
            
            
            // Compute solution statistics            
            
            Integer maxCluster = 0;
            for(Graph c : resultComponents){
                if(c.getVertexCount() > maxCluster){
                    maxCluster = c.getVertexCount();
                }
            }
            int[] ClusterSizes = new int[maxCluster];  
            
            
            Integer numberOfClusters = resultComponents.size();
            Set<Graph> blackMonochromaticClusters = new HashSet<>();
            Set<Graph> whiteMonochromaticClusters = new HashSet<>();
            Set<Graph> blackDominatedClusters = new HashSet<>();
            Set<Graph> whiteDominatedClusters = new HashSet<>();
            Set<Graph> singletons = new HashSet<>();
            Set<Graph> isolatedEdges = new HashSet<>();

            for(Graph c : resultComponents){
                if(!c.isCluster()){
                    System.out.println("!!!!!Not a cluster!!!!!");
                }
                else{
                    
                    ClusterSizes[c.getVertexCount()-1]++;
                    
                    if(c.isSingleton()){
                        singletons.add(c);
                    }
                    else if(c.isIsolatedEdge()){
                        isolatedEdges.add(c);
                    }
                    else if(c.isMonochromaticBlackCluster()){
                        blackMonochromaticClusters.add(c);
                    }
                    else if(c.isMonochromaticWhiteCluster()){
                        whiteMonochromaticClusters.add(c);
                    }
                    else if(c.isValidBlackDominatedCluster()){
                        blackDominatedClusters.add(c);
                    }
                    else if(c.isValidWhiteDominatedCluster()){
                        whiteDominatedClusters.add(c);
                    }
                    else System.out.println("No applicable cluster type");
                }               
            }
            
            Integer numberOfBlackMonochromaticClusters = blackMonochromaticClusters.size();
            Integer numberOfWhiteMonochromaticClusters = whiteMonochromaticClusters.size();
            Integer numberOfBlackDominatedClusters = blackDominatedClusters.size();
            Integer numberOfWhiteDominatedClusters = whiteDominatedClusters.size();
            Integer numberOfSingletons = singletons.size();
            Integer numberOfIsolatedEdges = isolatedEdges.size();
            
            
            
            // Set solution
            Set<Edge<K>> solutionSet = new HashSet<>();
            for(Edge<K> e : edgesInG){
                double edgeResult = edgeVariables.get(e).get(GRB.DoubleAttr.X);
                if(edgeResult <= 0.0001){
                    solutionSet.add(e);
                }    
            }
            
            Integer numberOfMods = solutionSet.size();
            
            
            String clusterhistogramm = Arrays.toString(ClusterSizes);
   
      
      //      model.write("BCD.lp");
      //      model.write("BCD.attr");
            
            // Dispose of model and environment
            model.dispose();
            env.dispose();
            
            
            // Conctructing Result Output            
            String result = Integer.toString(blackVerticesInG.size()) + "\t" +
                            Integer.toString(whiteVerticesInG.size()) + "\t" +
                            Integer.toString(edgesInG.size()) + "\t" +
                            finishedInTime + "\t" +
                            Long.toString(convertedILPtime) + "\t" +
                            numberOfMods.toString() + "\t" +                    
                            numberOfClusters.toString() + "\t" +
                            numberOfSingletons.toString() + "\t" +
                            numberOfIsolatedEdges.toString() + "\t" +
                            numberOfBlackMonochromaticClusters.toString() + "\t" +
                            numberOfWhiteMonochromaticClusters.toString() + "\t" +
                            numberOfBlackDominatedClusters.toString()+ "\t" +
                            numberOfWhiteDominatedClusters.toString()+ "\t" +
                            clusterhistogramm + "\t" 

                             ;
                               
            
            System.out.println(solutionSet.size()+" edge mods: "+solutionSet);
            
            return new Tuple(solutionSet,result);
            
        } catch (GRBException e) {
            System.out.println(e.getMessage());
            return new Tuple(new HashSet<>(),"");
        }
        catch (IOException e){
            System.out.println(e.getMessage());
            return new Tuple(new HashSet<>(),"");
        }
    }
    
    
    
    
    /*
    * Solves BCE for given graph G and returns a tuple of the solution set and statistics string
    */
    public Tuple<Set<Edge<K>> , String> solveBCE(Graph G) {
       
        try {
            String path =  System.getProperty("java.library.path");
            System.out.println("Java path: " + path);
            
            long StartTime = System.nanoTime();
            
            //setup
            GRBEnv    env   = new GRBEnv();
            FileWriter logfile = new FileWriter("solveBCE.log");
            GRBModel  model = new GRBModel(env);
            
            model.set(GRB.DoubleParam.TimeLimit, 600);
            model.set(GRB.IntParam.LazyConstraints, 1);
            
                                    
            Set<Edge<K>> edgesInG = G.getEdges();         
            Set<Edge<K>> missingEdgesInG = G.getMissingEdges();
            Set<Vertex<K>> verticesInG = G.getVertices();
            Set<Vertex<K>> blackVerticesInG = G.getBlackVertices();
            Set<Vertex<K>> whiteVerticesInG = G.getWhiteVertices();

            // Create edge variables
            HashMap<Edge<K>,GRBVar> edgeVariables = new HashMap<>();
            for(Edge<K> e : edgesInG){
               edgeVariables.put(e,model.addVar(0.0, 1.0, 0.0, GRB.BINARY, e.toString())); 
            }
            HashMap<Edge<K>,GRBVar> missingEdgeVariables = new HashMap<>();
            for(Edge<K> e : missingEdgesInG){
               missingEdgeVariables.put(e,model.addVar(0.0, 1.0, 0.0, GRB.BINARY, e.toString())); 
            }
            
            // Set objective
            GRBLinExpr objective = new GRBLinExpr();
            for(GRBVar var : edgeVariables.values()){
                objective.addTerm(-1.0, var);
            }
            for(GRBVar var : missingEdgeVariables.values()){        
                objective.addTerm(1.0,var);
            }
            objective.addConstant(edgesInG.size());
            model.setObjective(objective, GRB.MINIMIZE);
            
            
            model.update();
              

            // Add constraints for each P3
            int P3ConstraintCount = 0;
            
            Set<Set<Edge<K>>> p3s = G.getP3s();
            for(Set<Edge<K>> p3 : p3s){
                Set<Vertex<K>> verticesInP3s = new HashSet<>();
                for(Edge<K> p : p3){
                    verticesInP3s.addAll(p.getNodes());
                }
                Iterator vertexIterator = verticesInP3s.iterator();
                Vertex<K> u = (Vertex<K>) vertexIterator.next();
                Vertex<K> v = (Vertex<K>) vertexIterator.next();
                Vertex<K> w = (Vertex<K>) vertexIterator.next();
                Edge<K> uv = new Edge(u,v);
                Edge<K> vw = new Edge(v,w);
                Edge<K> uw = new Edge(u,w);
                Set<Edge<K>> edges = new HashSet<>();
                edges.add(uv);
                edges.add(vw);
                edges.add(uw);
                for(Edge<K> e : edges){
                    GRBLinExpr term = new GRBLinExpr();
                    for(Edge<K> e2 : edges){
                        if(p3.contains(e2)){
                            if(e2.equals(e)){
                                term.addTerm(-1.0,edgeVariables.get(e2));
                            }
                            else term.addTerm(1.0,edgeVariables.get(e2));
                        }
                        else if(e2.equals(e)){
                            term.addTerm(-1.0,missingEdgeVariables.get(e2));
                            }
                            else term.addTerm(1.0,missingEdgeVariables.get(e2));
                    } 
                    model.addConstr(term, GRB.LESS_EQUAL, 1.0, p3.toString());
                    P3ConstraintCount ++;
                }   
                
            }
            System.out.println("P3 Constraints added: "+P3ConstraintCount);
            
            
            // Add constraints for each triangle (potential P3)
            int TriangleConstraintCount = 0;
            Set<Set<Edge<K>>> triangles = G.getTriangles();
            for(Set<Edge<K>> tri : triangles){
                for(Edge<K> e : tri){
                    GRBLinExpr term = new GRBLinExpr();
                    for(Edge<K> e2 : tri){
                        if(e2.equals(e)){
                            term.addTerm(-1.0,edgeVariables.get(e2));
                        }
                        else term.addTerm(1.0,edgeVariables.get(e2));                      
                    } 
                    model.addConstr(term, GRB.LESS_EQUAL, 1.0, tri.toString());
                    TriangleConstraintCount ++;
//                    System.out.println("Triangle constraint added: "+ 
//                        term.getCoeff(0)+term.getVar(0).get(GRB.StringAttr.VarName)+" "+
//                        term.getCoeff(1)+term.getVar(1).get(GRB.StringAttr.VarName)+" "+
//                        term.getCoeff(2)+term.getVar(2).get(GRB.StringAttr.VarName));
                }                                 
            }
            System.out.println("Triangle Constraints added: "+TriangleConstraintCount);
            
            
            
            //Add lazy constraints via callback
            CallbackBCE cb = new CallbackBCE(model.getVars(), logfile, G, edgeVariables, missingEdgeVariables);
            model.setCallback(cb);
            
            
            
            // Optimize model
            model.optimize();
            
            
            
            
            long FinishTime = System.nanoTime();
            long ILPTime = FinishTime - StartTime;
            long convertedILPtime = TimeUnit.MILLISECONDS.convert(ILPTime, TimeUnit.NANOSECONDS);
            
            
            
            // Construction of result graph 
            Graph resultGraph = new Graph(verticesInG);
            for(Edge e : edgeVariables.keySet()){
                if((edgeVariables.get(e).get(GRB.DoubleAttr.X))>=0.9999){
                    Vertex<K>[] nodeArray = e.getNodeArray();
                    resultGraph.addEdge(nodeArray[0],nodeArray[1]);
                }
            }
            for(Edge e : missingEdgeVariables.keySet()){
                if((missingEdgeVariables.get(e).get(GRB.DoubleAttr.X))>=0.9999){
                    Vertex<K>[] nodeArray = e.getNodeArray();
                    resultGraph.addEdge(nodeArray[0],nodeArray[1]);
                }                 
            }
            

            
            Integer deletionCount = 0;
            Integer insertionCount = 0;
            
            // Set solution and calculate number of deletions/insertions
            Set<Edge<K>> solutionSet = new HashSet<>();
            for(Edge<K> e : edgesInG){
                double edgeResult = edgeVariables.get(e).get(GRB.DoubleAttr.X);
                if(edgeResult <= 0.0001){
                    solutionSet.add(e);
                    deletionCount++;
                }    
            }
            for(Edge<K> e : missingEdgesInG){
                double missingEdgeResult = missingEdgeVariables.get(e).get(GRB.DoubleAttr.X);
                if(missingEdgeResult >= 0.9999){
                    solutionSet.add(e);
                    insertionCount++;
                }    
            }
        
           
            // Compute solution statistics
            
            String finishedInTime;
            if(model.get(GRB.IntAttr.Status)==(GRB.Status.TIME_LIMIT)){
                finishedInTime = "no";
            }
            else finishedInTime = "yes";
            
            
            Set<Graph> resultComponents = resultGraph.getConnectedComponents();
            
            
            
            Integer maxCluster = 0;
            for(Graph c : resultComponents){
                if(c.getVertexCount() > maxCluster){
                    maxCluster = c.getVertexCount();
                }
            }
            int[] ClusterSizes = new int[maxCluster];
            
            
            
            Integer numberOfClusters = resultComponents.size();
            Set<Graph> blackMonochromaticClusters = new HashSet<>();
            Set<Graph> whiteMonochromaticClusters = new HashSet<>();
            Set<Graph> blackDominatedClusters = new HashSet<>();
            Set<Graph> whiteDominatedClusters = new HashSet<>();
            Set<Graph> singletons = new HashSet<>();
            Set<Graph> isolatedEdges = new HashSet<>();
            
            
            Set<Graph> noClusters = new HashSet<>();
            Set<Graph> notValidClusters = new HashSet<>();
            
            for(Graph c : resultComponents){
                if(!c.isCluster()){
                    noClusters.add(c);
                    System.out.println("!!!!!NOT A CLUSTER!!!!!");
                }
                else{
                    
                    ClusterSizes[c.getVertexCount()-1]++;
                    
                    if(c.isSingleton()){
                        singletons.add(c);
                    }
                    else if(c.isIsolatedEdge()){
                        isolatedEdges.add(c);
                    }
                    else if(c.isMonochromaticBlackCluster()){
                        blackMonochromaticClusters.add(c);
                    }
                    else if(c.isMonochromaticWhiteCluster()){
                        whiteMonochromaticClusters.add(c);
                    }
                    else if(c.isValidBlackDominatedCluster()){
                        blackDominatedClusters.add(c);
                    }
                    else if(c.isValidWhiteDominatedCluster()){
                        whiteDominatedClusters.add(c);
                    }
                    else {
                        notValidClusters.add(c);
                        System.out.println("!!!! No applicable cluster type  !!!");
                    }
                }               
            }
            
            Integer numberOfBlackMonochromaticClusters = blackMonochromaticClusters.size();
            Integer numberOfWhiteMonochromaticClusters = whiteMonochromaticClusters.size();
            Integer numberOfBlackDominatedClusters = blackDominatedClusters.size();
            Integer numberOfWhiteDominatedClusters = whiteDominatedClusters.size();
            Integer numberOfSingletons = singletons.size();
            Integer numberOfIsolatedEdges = isolatedEdges.size();
            
            Integer numberOfNotClusters = noClusters.size();
            Integer numberOfNotValidClusters = notValidClusters.size();
                    
            
            String clusterhistogramm = Arrays.toString(ClusterSizes);
   
      
      //      model.write("BCE.lp");
      //      model.write("BCE.attr");
            
            // Dispose of model and environment
            model.dispose();
            env.dispose();
            
            
            // Conctructing Result Output            
            String result = Integer.toString(blackVerticesInG.size()) + "\t" +
                            Integer.toString(whiteVerticesInG.size()) + "\t" +
                            Integer.toString(edgesInG.size()) + "\t" +
                            finishedInTime + "\t" +
                            Long.toString(convertedILPtime) + "\t" +            
                            deletionCount + "\t" +
                            insertionCount + "\t" + 
                            numberOfClusters.toString() + "\t" +
                            numberOfSingletons.toString() + "\t" +
                            numberOfIsolatedEdges.toString() + "\t" +
                            numberOfBlackMonochromaticClusters.toString() + "\t" +
                            numberOfWhiteMonochromaticClusters.toString() + "\t" +
                            numberOfBlackDominatedClusters.toString()+ "\t" +
                            numberOfWhiteDominatedClusters.toString()+ "\t" +
                            clusterhistogramm + "\t" +
                            numberOfNotClusters.toString() + "\t" +
                            numberOfNotValidClusters.toString()+ "\t"

                             ;
            
            
            System.out.println(solutionSet.size()+" edge mods: "+solutionSet);
            return new Tuple(solutionSet,result);
            
        } catch (GRBException e) {
            System.out.println(e.getMessage());
            return new Tuple(new HashSet<>(),"");
        }
        catch (IOException e){
            System.out.println(e.getMessage());
            return new Tuple(new HashSet<>(),"");
        }
    }
    
}